package TreeSnatcher.Core;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.image.BufferedImage;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Vector;

import TreeSnatcher.GUI.GUIActions;
import TreeSnatcher.GUI.ImageBuffer;
import TreeSnatcher.GUI.ImagePanel;
import TreeSnatcher.GUI.Wizard;
import TreeSnatcher.Utils.NumberUtility;
//import TreeSnatcher.Utils.UniqueColorSource;

public class TreeTopology implements Constants
{
	private static final long serialVersionUID = 1L;
	private ImageOperations imageOperations;
	private ImagePanel imagePanel;
	private ImageBuffer imageBuffer;
	private Wizard wizard;
	private NewickCalculator newickCalculator;
	private MainWindow mainWindow;
	//private UniqueColorSource uniqueColorSource;
	Graphics2D g2;
	BufferedImage binarizedImage;
	Vector<TreeNode> nodes = new Vector<TreeNode>();
	Vector<TreeNode> innerNodes = new Vector<TreeNode>();
	Vector<TreeNode> tips = new Vector<TreeNode>();
	Vector<Branch> branches = new Vector<Branch>();
	HashMap<Point, HashMap<Point, Integer>> distances;
	Vector<Branch> tempBranches = new Vector<Branch>();
	
	static ObjectOutputStream os = null;
	static ByteArrayOutputStream  out = null;
	
	public TreeTopology(ImageOperations iop, ImagePanel ip, ImageBuffer ib, Wizard wiz, MainWindow mainWindow, GUIActions guiActions)
	{
		this.imageOperations = iop;
		this.imagePanel = ip;
		this.imageBuffer = ib;
		this.wizard = wiz;
		this.mainWindow = mainWindow;
		newickCalculator = new NewickCalculator(nodes, branches, ip, wizard, this, imageOperations);
		binarizedImage = imageBuffer.getBinarizedImage();
	}
	
	public void collectTips(int fgColor)
	{
		Vector<Point> points = imageOperations.getNNeighborPixels(imagePanel.getCurrentImage(), 1, wizard.wholeImage, fgColor);
		for (int i = 0; i < points.size(); i++)
		{
			Point p = (Point) points.elementAt(i);
			Tip t = new Tip(p.x, p.y, false);
			//UniqueColorSource.addKey(t.getId());
			tips.add(t);
		}
		imagePanel.repaint();
		cleanupNodes(tips);
		points = null;
	}
	
	public void collectNodes(int fgColor)
	{
		// Collect candidate locations for inner nodes and tips
		Vector<Point> points = null, temp = new Vector<Point>();
		Point relPoint, nextPoint;
		points = imageOperations.getNNeighborPixels(imagePanel.getCurrentImage(), 3, wizard.wholeImage, fgColor);
		points.addAll(imageOperations.getNNeighborPixels(imagePanel.getCurrentImage(), 4, wizard.wholeImage, fgColor));
		
		do
		{
			// Store relPoint in temp
			relPoint = points.firstElement();
			temp.add(relPoint);
			
			// Find all direct neighbors of relPoint, store them in temp
			for (int j = 0; j < points.size(); j++)
			{
				nextPoint = points.elementAt(j);
				if ((!relPoint.equals(nextPoint) && (relPoint.distance(nextPoint) < 2.5)))
				{ 
					temp.add(nextPoint);
				}
			}
			
			int size = temp.size();
			//System.out.println("******");
			//for (int l = 0; l < temp.size(); l++) System.out.println(""+temp.elementAt(l).getLocation());
			
			if ((size == 1)) // A single candidate point or two candidates
			{
				//System.out.println("Single Point, size: "+temp.firstElement().getLocation());
				Point p = (Point) temp.firstElement();
				InnerNode in = new InnerNode(p.x, p.y, false);
				innerNodes.add(in);
			}
			else // Three or more neighboring candidate points
			{
				double sumX = 0, sumY = 0, meanX, meanY;
				
				for (int k = 0; k < size; k++)
				{
					sumX = sumX + temp.elementAt(k).x;
					sumY = sumY + temp.elementAt(k).y;
				}
				meanX = Math.ceil(sumX/size);
				meanY = Math.ceil(sumY/size);
				
				 // Select the point that is nearest to location (meanX, meanY)
				int shade = imagePanel.getCurrentImage().getRGB((int) meanX, (int) meanY);
				if ((shade == BIN1) || (shade == wizard.treeFloodColor)) // There is an ideal foreground pixel
				{					
					InnerNode in = new InnerNode((int) meanX, (int) meanY, false);							

				innerNodes.add(in);
					
				}				
				else // The ideal pixel is a background pixel				
				{
					// Instead, accept the first neighbor as inner node
					Point p = temp.firstElement();
					InnerNode in = new InnerNode(p.x, p.y, false);
					innerNodes.add(in);
				}
			}
			points.removeAll(temp);
			points.trimToSize();
			temp.clear();
		}
		while (!points.isEmpty());
		
		points = null;
		cleanupNodes(innerNodes);
		imagePanel.repaint();
	}
	
	public void wipeOutDoublettes()
	{
		// Remove those newly set nodes that are too near to nodes the the user created previously
		Vector<TreeNode> userCreatedNodes = new Vector<TreeNode>();
		// Collect user created nodes
		for (int i = 0; i < nodes.size(); i++)
		{
			TreeNode node = (TreeNode) nodes.elementAt(i);
			if (node.getUserGenerated()) userCreatedNodes.add(node);
		}

		// Compare user created nodes and all other nodes; delete those which are too near to user created nodes
		Iterator<TreeNode> nodesIterator = nodes.iterator();
		while (nodesIterator.hasNext())
		{
			TreeNode node = (TreeNode) nodesIterator.next();
			if (node.getUserGenerated() == false) // Node is program created
			{
				Iterator<TreeNode> ucnIterator = userCreatedNodes.iterator();
				while (ucnIterator.hasNext())
				{
					TreeNode userCreatedNode = (TreeNode) ucnIterator.next();
					
					if (userCreatedNode.dist2(node) <= 2) // Nodes are too near
					{
						if (node instanceof InnerNode) 
						{ 
							innerNodes.remove(node); 
							nodes.remove(node); 
						}
						else if (node instanceof Tip) 
						{ 
							tips.remove(node); 
							nodes.remove(node); 
						}
					}
				}
			}
		}
		userCreatedNodes = null;
	}
	
	
	public boolean collectBranches()
	{
		boolean couldBuildTopology = false;;
		tempBranches.clear();
		
		// Ensure that there are InnerNode- and Tip-objects
		Iterator<TreeNode> tit = nodes.iterator();
		int is = 0, os = 0;
		while (tit.hasNext())
		{
			TreeNode node = tit.next();
			if (node instanceof InnerNode) is++;
			else os++;
		}
		
		if ((is >= 1) && (os >= 2)) // For a sensible topology, there must be at least one inner node and two tips
		{
			// Color the node locations as they need to be distinguished from other foreground pixels
			imageOperations.colorNodePositions(nodes, Color.red);
			
			try
			{
				// Build the distance data structures for each node
				// {Distances} contains the distances between path pixels and the node position from which the path emerges 
				distances = new HashMap<Point, HashMap<Point, Integer>>();
				if (wizard.useElaboratePathFinding) imageOperations.buildDistancesElaborate(nodes, distances);
				else imageOperations.buildDistancesSimple(nodes, distances);
				
				// Follow the paths in a greedy fashion in reverse to find the paths with the shortest distances between the nodes
				Iterator<TreeNode> nodesIt = nodes.iterator();
	
				while (nodesIt.hasNext())
				{
					TreeNode seedNode = nodesIt.next();
					couldBuildTopology = imageOperations.buildPath(seedNode, nodes, distances, tempBranches);
				}
				
				if (couldBuildTopology)
				{
					for (int i = 0; i < tempBranches.size(); i++)
					{
						Branch b = tempBranches.elementAt(i);
						if (((b.getFirstNode() != null) && (b.getSecondNode() != null)) 
						&& (!b.getFirstNode().equals(b.getSecondNode()))) 
						{
							if ((b.getFirstNode() instanceof Tip) || (b.getSecondNode() instanceof Tip)) b.setOuterBranch(true);
							else b.setOuterBranch(false);
							addBranch(b);
						}
					}
					
					tempBranches.clear();
					
					if ((branches != null) && (branches.size() > 0))
					{
						// Get and use the inflection points on the branches
						imageOperations.identifyInflectionPoints(branches, nodes);
						imageOperations.determineBranchSegments(branches, nodes);
						imageOperations.calculateSegmentSlope(branches);
						imageOperations.determineLengthRelevantSegments(branches, nodes);
					}	
				}
			}
			catch (Exception ex) 
			{
				imagePanel.repaint();
			}
			
		}
		
		wizard.topologyDeterminationFinished = couldBuildTopology;
		imageOperations.colorNodePositions(nodes, Color.black);
		return couldBuildTopology;
	}
	
	public boolean foregroundPathExists(TreeNode n1, TreeNode n2)
	{
		// Get distance information for both nodes
		HashMap<Point, Integer> dist_n1 = distances.get(n1.getLocation());
		HashMap<Point, Integer> dist_n2 = distances.get(n2.getLocation());
		if ((dist_n1 != null) && (dist_n2 != null))
		{
			if ((dist_n1.containsKey(n2.getLocation())) && (dist_n2.containsKey(n1.getLocation()))) return true;
		}
		return false;
	}
	
	public int getPathLengthInPixels(TreeNode n1, TreeNode n2)
	{
		int length = -1;
		// Get distance data structures for both nodes
		HashMap<Point, Integer> dist_n1 = distances.get(n1.getLocation());
		HashMap<Point, Integer> dist_n2 = distances.get(n2.getLocation());
		if ((dist_n1 != null) && (dist_n2 != null))
		{
			length = (Integer) dist_n1.get(n2.getLocation());
		}
		return length;
	}
	
	
	public TreeNode getNodeAt(Point p)
	{
		TreeNode node = null;
		for (int i = 0; i < nodes.size(); i++)
		{
			node = (TreeNode) nodes.elementAt(i);
			if (node.getLocation().equals(p)) return node;
		}
		return null;
	}
	
	public Vector<Branch> getBranchesAtNode(TreeNode node)
	{
		Vector<Branch> selectedBranches = new Vector<Branch>();
		for (int i = 0; i < branches.size(); i++)
		{
			Branch branch = (Branch) branches.elementAt(i);
			if ((branch.getFirstNode() == node) || (branch.getSecondNode() == node)) selectedBranches.addElement(branch);
		}
		return selectedBranches;
	}
	
	public Vector<Branch> getInnerBranches() 
	{ 
		Vector<Branch> innerBranches = new Vector<Branch>();
		Iterator<Branch> it = branches.iterator();
		while (it.hasNext())
		{
			Branch b = it.next();
			if ((b.getFirstNode() instanceof InnerNode) && (b.getSecondNode() instanceof InnerNode)) innerBranches.add(b);
		}
		
		return innerBranches;
	}
	
	public Vector<Branch> getOuterBranches() 
	{ 
		Vector<Branch> outerBranches = new Vector<Branch>();
		Iterator<Branch> it = branches.iterator();
		while (it.hasNext())
		{
			Branch b = it.next();
			if ((b.getFirstNode() instanceof Tip) || (b.getSecondNode() instanceof Tip)) outerBranches.add(b);
		}
		
		return outerBranches;
	}
	
	public Branch getBranchBetweenNodes(TreeNode firstNode, TreeNode secondNode)
	{
		for (int i = 0; i < branches.size(); i++)
		{
			Branch branch = (Branch) branches.elementAt(i);
			if (((branch.getFirstNode() == firstNode) && (branch.getSecondNode() == secondNode))
			|| ((branch.getFirstNode() == secondNode) && (branch.getSecondNode() == firstNode))) 
				return branch;
		}
		return null;
	}
	
	public Vector<InnerNode> getInnerNodes()
	{
		Vector<InnerNode> innerNodes = new Vector<InnerNode>();
		Iterator<TreeNode> it = nodes.iterator();
		while (it.hasNext())
		{
			TreeNode treeNode = it.next();
			if (treeNode instanceof InnerNode) innerNodes.add((InnerNode) treeNode);
		}
		return innerNodes;
	}
	
	public Branch getBranch(int id)
	{
		Iterator<Branch> it = branches.iterator();
		while (it.hasNext())
		{
			Branch b = it.next();
			if (b.getId() == id) return b;
		}
		return null;
	}
	
	public boolean branchExists(Branch branch)
	{
		//Iterator<Branch> it = branches.iterator();
		if (branches.contains(branch)) return true;
		else return false;
	}
	
	public void moveNode(TreeNode node, Point p)
	{
		Iterator<TreeNode> nodesIt = nodes.iterator();
		while (nodesIt.hasNext())
		{ 
			TreeNode n = (TreeNode) nodesIt.next();
			if (n.equals(node))	n.setLocation(p);
		}
	}
	
	
	public void removeAllNodes()
	{
		if (nodes != null)
		{
			Iterator<TreeNode> nodesIt = nodes.iterator();
			while (nodesIt.hasNext())
			{ 
				TreeNode node = (TreeNode) nodesIt.next();
				if (node.getUserGenerated())
				{
					if  (wizard.mayDiscardUserGeneratedObjects) 
					{ 
						removeConnectedBranches(node);
						nodesIt.remove();
					}
				}
				else 
				{ 
					removeConnectedBranches(node);
					nodesIt.remove(); 
				}
			}
		}
		wizard.nodesCollected = false;
		wizard.referenceNodeId = -1;
		wizard.topologyDeterminationFinished = false;
		
		TreeNode.signalNode = -1;
		innerNodes.removeAllElements();
		tips.removeAllElements();
	}
	
	public void removeAllBranches()
	{
		if (branches != null)
		{
			Iterator<Branch> branchesIt = branches.iterator();
			while (branchesIt.hasNext())
			{ 
				Branch branch = (Branch) branchesIt.next();
				if (branch.getUserGenerated())
				{
					// Remove the branch object from the Vector
					// Use removeBranch() as the nodes' neighborship information must be updated
					if (wizard.mayDiscardUserGeneratedObjects) { removeBranch(branch); };
				}
				else { branchesIt.remove(); }; // Do not use removeBranch() as the nodes' neighborship information shall be preserved
			}
			wizard.referenceNodeId = -1;
			wizard.topologyDeterminationFinished = false;
			TreeNode.signalNode = -1;
		}
	}
	
	// Removes a node and its accompanying branches
	// The neighboring nodes' neighborship information gets adjusted (removeConnectedBranches)
	// Recalculate the tree topology
	public void removeNode(TreeNode node)
	{
		Iterator<TreeNode> nodesIt = nodes.iterator();
		while (nodesIt.hasNext())
		{ 
			if (node.equals((TreeNode) nodesIt.next()))
			{
				removeConnectedBranches(node);
				if (node.id == wizard.referenceNodeId) wizard.referenceNodeId = -1;
				nodesIt.remove(); 
			}
		}
		if (nodes.isEmpty()) { wizard.topologyDeterminationFinished = false;  TreeNode.signalNode = -1; };
		redoTopology();
	}
	
	// Removes the nodes within current rectangular selection and their accompanying branches
	// The neighboring nodes' neighborship information gets adjusted (removeConnectedBranches)
	// Recalculate the tree topology
	public void removeNodes(Vector<TreeNode> nodesInSelection)
	{
		for (int i = 0; i < nodesInSelection.size(); i++)
		{
			TreeNode nodeInSelection = (TreeNode) nodesInSelection.elementAt(i);
			Iterator<TreeNode> nodesIt = nodes.iterator();
			while (nodesIt.hasNext()) 
			{ 
				if (nodeInSelection.equals((TreeNode) nodesIt.next()))
				{
					removeConnectedBranches(nodeInSelection);
					nodesIt.remove();
				}
			}
		}
		if (nodes.isEmpty()) { wizard.topologyDeterminationFinished = false;  TreeNode.signalNode = -1; };
		redoTopology();
	}
	
	// Removes a branch
	// Purges for its two defining nodes the mutual reference in their neighborship objects
	// Recalculates the tree topology
	public void removeBranch(Branch branch)
	{
		Iterator<Branch> branchesIt = branches.iterator();
		while (branchesIt.hasNext()) 
		{ 
			if (branch.equals((Branch) branchesIt.next()))
			{
				TreeNode firstNode = branch.getFirstNode();
				Vector<TreeNode> firstNeighbors = firstNode.getNeighbors();
				TreeNode secondNode = branch.getSecondNode();
				Vector<TreeNode> secondNeighbors = secondNode.getNeighbors();
				
				if (firstNode.id == wizard.referenceNodeId) wizard.referenceNodeId = -1;
				if (secondNode.id == wizard.referenceNodeId) wizard.referenceNodeId = -1;
				firstNeighbors.remove(secondNode);
				secondNeighbors.remove(firstNode);
				
				// Remove the very branch
				branchesIt.remove();
			}
		}
		
		if (branches.isEmpty()) { wizard.topologyDeterminationFinished = false; TreeNode.signalNode = -1; };
		redoTopology();
		NumberUtility.setNumFractionDigits(branches);
	}
	
	// Removes the branches within current rectangular selection and their accompanying nodes
	// Purges for their two defining nodes the mutual reference in their neighborship objects
	// Recalculates the tree topology
	public void removeBranches(Vector<Branch> branchesInSelection)
	{
		for (int j = 0; j < branchesInSelection.size(); j++)
		{
			Branch branchInSelection = (Branch) branchesInSelection.elementAt(j);
			Iterator<Branch> branchesIt = branches.iterator();
			while (branchesIt.hasNext()) 
			{ 
				if (branchInSelection.equals((Branch) branchesIt.next()))
				{
					TreeNode firstNode = branchInSelection.getFirstNode();
					TreeNode secondNode = branchInSelection.getSecondNode();
					
					if (firstNode.id == wizard.referenceNodeId) wizard.referenceNodeId = -1;
					if (secondNode.id == wizard.referenceNodeId) wizard.referenceNodeId = -1;
					firstNode.getNeighbors().remove(secondNode);
					secondNode.getNeighbors().remove(firstNode);
					
					// Remove the very branch
					branchesIt.remove(); 
					redoTopology();
				}
			}
		}
		
		if (branches.isEmpty()) { wizard.topologyDeterminationFinished = false;  TreeNode.signalNode = -1; };
		NumberUtility.setNumFractionDigits(branches);
	}
	
	// Remove the branches which are connected to this node
	private void removeConnectedBranches(TreeNode node)
	{
		TreeNode firstNode, secondNode;
		
		Iterator<Branch> branchesIt = branches.iterator();
		while (branchesIt.hasNext())
		{
			Branch branch = (Branch) branchesIt.next();
			firstNode = branch.getFirstNode();
			secondNode = branch.getSecondNode();
			if ((firstNode.equals(node)) || (secondNode.equals(node))) 
			{
				if (firstNode.id == wizard.referenceNodeId) wizard.referenceNodeId = -1;
				if (secondNode.id == wizard.referenceNodeId) wizard.referenceNodeId = -1;
				firstNode.getNeighbors().remove(secondNode);
				secondNode.getNeighbors().remove(firstNode);
		
				branchesIt.remove();
			}
		}
		
		if (branches.isEmpty()) { wizard.topologyDeterminationFinished = false; TreeNode.signalNode = -1; };
		NumberUtility.setNumFractionDigits(branches);
	}
	
	private void cleanupNodes(Vector<TreeNode> treeNodes)
	{
		// TODO: Determine dmin based on the local foreground in the BINARIZED (not thinned) image
		double dmin = 3;
		if (wizard.useElaboratePathFinding) dmin = 3;
		double dmin2 = dmin*dmin;
		Object[] nodeArray = treeNodes.toArray();
		Vector<TreeNode> goodNodes = new Vector<TreeNode>(treeNodes.size());
		for (int i = 0; i < nodeArray.length; i++)
		{
			if (nodeArray[i] != null)
			{
				TreeNode n1 = (TreeNode) nodeArray[i];
				goodNodes.add(n1);
				
				// Delete all remaining nodes close to n1
				for (int j=i+1; j<nodeArray.length; j++)
				{
					if (nodeArray[j] != null)
					{
						TreeNode n2 = (TreeNode) nodeArray[j];
						if (n1.dist2(n2) < dmin2) nodeArray[j] = null;
					}
				}
			}
		}
		treeNodes = null;
		nodeArray = null;
		nodes.addAll(goodNodes);
		goodNodes = null;
	}
	
	
	public InnerNode addInnerNode(Point l)
	{
		InnerNode  in = null;
		boolean insert = true;
		if (l != null) 
		{
			in = new InnerNode(l.x,l.y, true);
			
			if (nodes.size() > 0)
			{
				Iterator<TreeNode> nodesIt = nodes.iterator();
				
				while (nodesIt.hasNext())
				{
					TreeNode n = nodesIt.next();
					int x = n.getX();
					int y = n.getY();
					if ((Math.abs(x-l.x) < 2) && (Math.abs(y-l.y) < 2)) insert = false;
				}
			}
			if (insert) nodes.add(in);
		}
		
		if (wizard.topologyDeterminationFinished) redoTopology();
		return in;
	}
	
	public Tip addTip(Point l)
	{
		Tip t = null;
		boolean insert = true;
		if (l != null) 
		{
			t = new Tip(l.x, l.y, true);
			
			if (nodes.size() > 0)
			{
				Iterator<TreeNode> nodesIt = nodes.iterator();
				while (nodesIt.hasNext())
				{
					TreeNode n = nodesIt.next();
					int x = n.getX();
					int y = n.getY();
					if ((Math.abs(x-l.x) < 2) && (Math.abs(y-l.y) < 2)) insert = false;
				}
			}
			if (insert) nodes.add(t);
		}
		
		if (wizard.topologyDeterminationFinished) redoTopology();
		return t;
	}
	
	
	public void addBranch(Branch branch)
	{
		boolean duplicateBranch = false;
		boolean tipWithMultipleBranches = false;
		if (branch != null) 
		{
			// Do not add a branch that already exists
			TreeNode firstNode = branch.getFirstNode();
			TreeNode secondNode = branch.getSecondNode();
			for (int i = 0; i  < branches.size(); i++)
			{
				Branch branch2 = (Branch) branches.elementAt(i);
				TreeNode firstNode2 = branch2.getFirstNode();
				TreeNode secondNode2 = branch2.getSecondNode();
				
				// Avoid duplicate branches
				if (((firstNode.equals(firstNode2)) && (secondNode.equals(secondNode2))) 
				|| ((firstNode.equals(secondNode2)) && (secondNode.equals(firstNode2)))) duplicateBranch = true;
				
				// Avoid Tip with more than one emerging branch
				if 		((firstNode2 instanceof Tip) && (firstNode instanceof Tip) && (firstNode.equals(firstNode2))) tipWithMultipleBranches = true;
				else if ((firstNode2 instanceof Tip) && (secondNode instanceof Tip) && (secondNode.equals(firstNode2)))  tipWithMultipleBranches = true;
				else if ((secondNode2 instanceof Tip) && (firstNode instanceof Tip) && (firstNode.equals(secondNode2)))  tipWithMultipleBranches = true;
				else if ((secondNode2 instanceof Tip) && (secondNode instanceof Tip) && (secondNode.equals(secondNode2)))  tipWithMultipleBranches = true;
			}
			if ((!duplicateBranch) && (!tipWithMultipleBranches))
			{
				branches.add(branch);
				firstNode.addNeighbor(secondNode);
				secondNode.addNeighbor(firstNode);
			}
			
			if (!wizard.branchesCollecting) redoTopology(); // Only recalculate the topology after a manual branch addition
		}
	}
	
	public InnerNode insertRootAtBranch(Branch branch)
	{
		if (branch == null) return null;
		// Get the path of this branch
		Vector<Point> path = branch.getPath();
		
		// Go to the middle of the branch
		Point p = null;
		int halfway = (int) (0.5*(branch.getTotalLengthInPixels()+1));
		for (int i = 0; i < halfway; i++) p = path.elementAt(i);
		InnerNode in = new InnerNode(p, true);
		nodes.add(in);
		wizard.referenceNodeId = in.id;
		redoTopology();
		
		return in;
	}
	
	public void reset()
	{
		TreeNode.num = 0;
		Branch.num = 0;
		wizard.referenceNodeId = -1;
		
		nodes.removeAllElements();
		branches.removeAllElements();
		innerNodes.removeAllElements();
		tips.removeAllElements();
	}
	
	public Vector<TreeNode> getNodes()
	{
		if (nodes == null) return null;
		else return this.nodes;
	}
	
	public void setNodes(Vector<TreeNode> nodes)
	{
		this.nodes = nodes;
	}
	
	public Vector<Branch> getBranches()
	{
		if (branches == null) return null;
		else return this.branches;
	}
	
	public void setBranches(Vector<Branch> branches)
	{
		this.branches = branches;
	}
	
	// Returns the lowest pairwise distance over all TreeNode objects
	public int getLowestPairwiseDistance()
	{
		// Naive approach, better use a Sweep-line-approach
		double minDistance = Integer.MAX_VALUE;
		
		for (int i = 0; i < this.getNodes().size(); i++)
		{
			for (int j = 0; j < nodes.size(); j++)
			{
				Point p1 = new Point(nodes.elementAt(i).getLocation());
				Point p2 = new Point(nodes.elementAt(j).getLocation());
				
				if (!p1.equals(p2))
				{
					double dist = p1.distance(p2);
					if (dist < minDistance) minDistance = dist;
				}
			}
		}

		if (minDistance == Integer.MAX_VALUE) return 0;
		else return (int) minDistance;
	}
	
	
	public String transformIntoNewickExpression()
	{
		// Mark all branches as "not visited"
		for (int i = 0; i < branches.size(); i++)
		{
			((Branch) branches.elementAt(i)).setVisited(false);
		}
		// Mark all nodes as "not visited"
		for (int i = 0; i < nodes.size(); i++)
		{
			((TreeNode) nodes.elementAt(i)).setVisited(false);
		}
		
		if (nodes.size() == 0) return " The tree nodes need to be determined.";
		else if (branches.size() == 0) return "The branches need to be determined.";
		else 
		{	
			if ((wizard.topologyDeterminationFinished == true) && (wizard.mostRecentTopologyError != 0))
			{
				wizard.showResultFrame = true;
				mainWindow.swapWizardAndNewickPanels();
			}
			
			// Determine the Newick expression for the tree topology
			return newickCalculator.calculateNewickExpression(nodes, branches);
		}
	}
	
	// Redo the topology check, display it on-screen
	private void redoTopology()
	{
		transformIntoNewickExpression();
		imagePanel.updateResultFrame();
	}
	
	public void redisplayNewickExpression()
	{
		try
		{
			newickCalculator.calculateNewickExpression(nodes, branches);
			imagePanel.updateResultFrame();
		}
		catch (NullPointerException e) {System.out.println("Newick Expression: Exception");};
	}
	
	public String getNewickExpression()
	{
		return newickCalculator.getNewickString();
	}
	
	public void setTaxonName(int id, String name)
	{
		Iterator<TreeNode> nodesIt = nodes.iterator();
		
		while (nodesIt.hasNext())
		{
			TreeNode n = nodesIt.next();
			if (id == n.getId()) { ((Tip) n).setTaxonName(name); imagePanel.repaint(); }
		}
	}
	
	public boolean findNode(int id)
	{
		Iterator<TreeNode> nodesIt = nodes.iterator();
		
		while (nodesIt.hasNext())
		{
			TreeNode n = nodesIt.next();
			if (id == n.getId()) return true;
		}
		return false;
	}
	
	public void issueSnapshot(Vector<Branch> branches, Vector<TreeNode> nodes)
	{
		this.branches = branches;
		this.nodes = nodes;
		redoTopology();
	}
	
	public ByteArrayOutputStream getSerializedTopology()
	{
		// Store the current topology in memory
		ByteArrayOutputStream out = new ByteArrayOutputStream();
		ObjectOutputStream os = null;
		try 
		{
			os = new ObjectOutputStream(out);
			os.writeObject(nodes);
			os.writeObject(branches);
			os.writeObject(TreeNode.num);
			os.writeObject(TreeNode.signalNode);
			os.writeObject(Branch.num);
			os.writeObject(Tip.nameCnt);
			os.flush();
		} 
		catch (IOException e) {}
		return out;
	}
	
	
	/*
	public void collectBranches()
	{
		wizard.topologyDeterminationFinished = false;
		
		Vector<InnerNode> innerNodes = new Vector<InnerNode>();
		Vector<Tip> tips = new Vector<Tip>();
		Vector<Branch> newBranches = new Vector<Branch>();
			
		for (int i = 0; i < nodes.size(); i++)
		{
			// Keep InnerNode objects and Tip objects in separate containers
			TreeNode node = nodes.elementAt(i);
			if (node != null)
			{
				if (node instanceof Tip) tips.add((Tip) node);
				else if (node instanceof InnerNode) innerNodes.add((InnerNode) node);
			}
		}
		
		// Color the coordinates of all inner nodes in a unique signal color for mutual identification
		imageOperations.colorPixelsUniquely(innerNodes);
		//imagePanel.repaint(imagePanel.scrollPane.getVisibleRect());
		imagePanel.repaint();
		
		// Construct the tree's branches via tracing the image foreground from a tip to an inner node
		for (int i = 0; i < tips.size(); i++)
		{
			Tip tip = tips.elementAt(i);
			if (tip != null)
			{
				Branch branch = new Branch(tip, null, false);
				imageOperations.tracePathFromTip(tip, branch, innerNodes);
				if (((branch.getFirstNode() != null) && (branch.getSecondNode() != null)) 
				&& (!branch.getFirstNode().equals(branch.getSecondNode()))) 
				{
					addBranch(branch);
				}
			}
			imagePanel.repaint();
		}
		
		// Re-color the inner nodes in their signal colors
		imageOperations.colorPixelsUniquely(innerNodes);
		
		// Construct the tree's branches between inner nodes employing a similar technique
		Iterator<InnerNode> inIt = innerNodes.iterator();
		while (inIt.hasNext())
		{
			InnerNode innerNode = inIt.next();
			newBranches.clear();
			imageOperations.traceAllPathsFromInnerNode(innerNode, newBranches, innerNodes);
			
			// Validate the newly traced branches: Ignore doublets
			Iterator<Branch> nbIt = newBranches.iterator();
			while (nbIt.hasNext())
			{
				Branch branch = nbIt.next();
				if (branch != null)
				{
					if ((branch.getFirstNode() != null) && (branch.getSecondNode() != null) 
					&& (!branch.getFirstNode().equals(branch.getSecondNode()))) 
					{
						addBranch(branch);
					}
				}
				imagePanel.repaint();
			}
		}
		
		// Restore all foreground shades, background has shade BIN0
		// TODO: Reicht das so? Bei nicht binarisierten Bildern wird das ganze Bild schwarz, darf nicht sein...
		if (wizard.isBinarized) imageOperations.blackenFGPixels(imagePanel.getCurrentImage(), BIN0, wizard.wholeImage);
		
		// 2nd pass: Calculate the lengths in pixels of each path
		determineLengths();
		wizard.topologyDeterminationFinished = true;
	}
	
	private void determineLengths()
	{
		imagePanel.stopProgressPane();
		BufferedImage image = imageBuffer.cloneProcessedImage(); // Grab current image state from ImagePanel
		//new DebugFrame(image, "Calculated Paths");
		Iterator<Branch> it = branches.iterator();
		Vector<Point> oldPath = new Vector<Point>();
		Point p = null;
		
		while (it.hasNext())
		{
			Branch branch = it.next();
			if (branch != null) 
			{
				int length = imageOperations.calculatePathLength(branch, oldPath, image);
				if (length > 0) branch.setLengthInPixels(length);
			}
			
			// Blacken old path
			if (oldPath.size() > 0)
			{
				Iterator<Point> pit = oldPath.iterator();
				while (pit.hasNext()) 
				{
					p = pit.next();
					image.setRGB(p.x, p.y, BIN1);
				}
				oldPath.removeAllElements();
			}
		}
	}
	*/
	/*
	// Segregates tiny outer branches
	// Corrects the topology if it is not ambiguous
	// TODO: Drfen natrlich nur branches mit tips sein!
	public void segregateTinyBranches()
	{
		TreeNode nodeToErase = null, tip = null;						// The inner node of a tiny branch
		TreeNode node1, node2, outerNode1 = null, outerNode2 = null;	// The nodes that define the tiny branch
		Vector<TreeNode> neighbors;
		Vector<Branch> participatingBranches;
		Branch aBranch, branch1, branch2;
		
		int length = wizard.tinyBranchLength;
		
		for (int i = 0; i < branches.size(); i++)
		{
			Branch branch = (Branch) branches.elementAt(i);
			if ((branch != null) && (branch.getLength() <= length) && (branch.getLength() != 0.0d))
			{
				// Segregate the branch
				// 1. Identify the TreeNode objects involved 
				node1 = (TreeNode) branch.getFirstNode();
				node2 = (TreeNode) branch.getSecondNode();
				
				// Make sure than one node is a tip and one is an inner node
				if ((((node1 instanceof InnerNode) && (node2 instanceof InnerNode))
					|| ((node2 instanceof Tip) && (node1 instanceof Tip)))) break;
				
				if (node1 instanceof InnerNode) { nodeToErase = node1; tip = node2; }
				else { nodeToErase = node2; tip = node1; }
				
				neighbors = (Vector<TreeNode>) nodeToErase.getNeighbors();
				if (neighbors.size() == 3)
				{
					participatingBranches = getBranchesWithNode(nodeToErase);
					// Among those delete the branch that equals the tiny branch itself
					for (int  j= 0; j < participatingBranches.size(); j++)
					{
						aBranch = (Branch) participatingBranches.elementAt(j);
						if (aBranch == branch) { participatingBranches.removeElement(aBranch); j--; }
					}
					
					// The nodes that the remaining branches have not in common
					// are the "outer" nodes of those branches
					branch1 = (Branch) participatingBranches.elementAt(0);
					branch2 = (Branch) participatingBranches.elementAt(1);
					
					if ((branch1.getFirstNode() == branch2.getFirstNode()) ||
					(branch1.getFirstNode() == branch2.getSecondNode())) outerNode1 = branch1.getSecondNode();
					if ((branch1.getSecondNode() == branch2.getFirstNode()) ||
					(branch1.getSecondNode() == branch2.getSecondNode())) outerNode1 = branch1.getFirstNode();
					
					if ((branch2.getFirstNode() == branch1.getFirstNode()) ||
					(branch2.getFirstNode() == branch1.getSecondNode())) outerNode2 = branch2.getSecondNode();
					if ((branch2.getSecondNode() == branch1.getFirstNode()) ||
					(branch2.getSecondNode() == branch1.getSecondNode())) outerNode2 = branch2.getFirstNode();
					
					// Erase the tiny branch only if this would not induce a link between two tips
					if ((outerNode1 instanceof InnerNode) && (outerNode2 instanceof InnerNode))
					{
						// Erase the objects that are no longer used
						nodes.removeElement(tip);
						tips.removeElement(tip);
						branches.remove(branch);
						branches.remove(branch1);
						branches.remove(branch2);
						
						// Purge nodeToErase as a neighbor from the "outer" nodes
						outerNode1.getNeighbors().remove(nodeToErase);
						outerNode2.getNeighbors().remove(nodeToErase);
						
						// Remove nodeToErase
						nodes.remove(nodeToErase);
						innerNodes.remove(nodeToErase);
						
						// Add the branch that links the outer nodes to the topology
						Branch newBranch = new Branch(outerNode1, outerNode2, false);
						addBranch(newBranch);
					}
				}
				else if (neighbors.size() > 3) // No topology change
				{
					nodes.removeElement(tip);
					nodeToErase.getNeighbors().removeElement(tip);
					branches.removeElement(branch);
				}
			}
		}
		imagePanel.repaint();
	}
	*/
	
	/*
	public int determineEdgeWidth(Point2D referencePoint)
	{
		// Determine the width of the edge on which the user sets the reference point
		// Examine the surroundings of this location in the binarized (but not skeletonized) image
		// The skeleton of the edge is situated in the middle of the former edge,
		// therefore the diameter of the edge can be estimated by inscribing
		// a circle of maximal diameter 
		
		BufferedImage source = getSourceImage();
		//Point2D origin = TreeSnatcher.workView.getGlassPaneViewportOrigin();
		//int ox = (int) origin.getX();
		//int oy = (int) origin.getY();
		double x = referencePoint.getX();
		double y = referencePoint.getY();
		double[] currentCoords;
		double c1x, c1y, c2x, c2y, c3x, c3y;
		int r = 1;
		PathIterator pathIterator;
		int segmentType;
		boolean maxRadiusFound = false;
		boolean notOnEdge = false;
		Ellipse2D ellipse;
		int bgCol = TreeSnatcher.imageBuffer.getBackgroundColor().getRGB();

		do
		{
			ellipse = new Ellipse2D.Double(x-r, y-r, r, r); // Ellipse at (x, y) with radius r
			pathIterator = ellipse.getPathIterator(null);
			notOnEdge = false;
			
			do
			{	
				pathIterator.next();
				currentCoords = new double[6];
				if (!pathIterator.isDone()) // Stay on the ellipse's boundary
				{
					pathIterator.currentSegment(currentCoords);
					c1x = currentCoords[0];
					c1y = currentCoords[1];
					// Restrict coordinates to the image's boundary
					if (c1x < 0) c1x = 0;
					if (c1x > source.getWidth()) c1x = source.getWidth();
					if (c1y < 0) c1y = 0;
					if (c1y > source.getHeight()) c1y = source.getHeight();
					int c = source.getRGB((int) c1x, (int) c1y);
					//System.out.println("Coords: "+c1x+", "+c1y+" Color: "+new Color(c));
					if ((c == bgCol) && (c1x != 0) && (c1y != 0)) notOnEdge = true;
				}
			}
			while (!(pathIterator.isDone()));
			
			if (notOnEdge == true) maxRadiusFound = true;
			else r++;
		}
		while (maxRadiusFound == false);
		
		edgeWidth = (int) Math.max(1, Math.ceil(2*r))+1;
		//edgeWidth = (int) Math.max(2, Math.floor(2*r));
		//System.out.println("Edge width: "+edgeWidth);
	}
	*/
	
	/*
	public boolean storeSnapshot()
	{
		File file = new File(Preferences.snapshotPath + File.separator + "snapshot1" + ".ser");
		try
		{
			FileOutputStream fs = new FileOutputStream(file);
			ObjectOutputStream os = new ObjectOutputStream(fs);
			os.writeObject(branches);
			os.close();
		}
		catch (IOException e)
		{
			System.out.println("Could not write Branches");
		}
		return true;
		
	}
	
	public boolean getSnapshot()
	{
		File file = new File(Preferences.snapshotPath + File.separator + "snapshot1" + ".ser");
		
		try
		{
			FileInputStream fs = new FileInputStream(file);
			ObjectInputStream is = new ObjectInputStream(fs);
			branches = (Vector<Branch>) is.readObject();
			System.out.println(""+branches.toString());
			is.close();
		}
		catch (ClassNotFoundException e) 
		{
			System.out.println("Class not found.");
		}
		catch (IOException e)
		{
			System.out.println("Could not read Branches");
		}
		return true;
	}
	*/
}
